package pa;

import java.nio.IntBuffer;

import org.lwjgl.BufferUtils;
import org.lwjgl.PointerBuffer;
import org.lwjgl.opencl.CL10;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLProgram;

import pa.cl.CLUtil;
import pa.cl.CLUtil.PlatformDevicePair;
import pa.cl.OpenCL;
import pa.util.IOUtil;

public class Reduction
{
    
    private static CLContext context;
    private static CLKernel kernel;
    private static CLMem input;
    private static CLCommandQueue queue;
    private static CLProgram program;
    
    public static void runTest13(int expo, int localWorkSize)
    {
        int size = 1 << expo;
         
        localWorkSize = 7;
        
        System.out.println(String.format("--- Running test size=%d (%d), lws=%d", size, expo, localWorkSize));
        
        if(size/2 < localWorkSize || size % localWorkSize != 0)
        {
            System.out.println("    size/2 < localWorkSize || size % localWorkSize != 0, this case is not handled yet\n");
            return;
        }
        
        IntBuffer data = BufferUtils.createIntBuffer(size);
        for(int i = 0; i < size; ++i)
        {
            data.put(1 + i);
        }
        
        data.position(0);
        
        input = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_WRITE | CL10.CL_MEM_COPY_HOST_PTR, data);
        
        OpenCL.clSetKernelArg(kernel, 0, input);
        OpenCL.clSetKernelArg(kernel, 1, 1);
        
        PointerBuffer workSize = new PointerBuffer(1);
        workSize.put(0, size / 2);
        
        PointerBuffer workSizeLocal = new PointerBuffer(1);
        workSizeLocal.put(0, localWorkSize);
        
        int s = (int)workSize.get(0);
        
        int log = 1 + (int)(Math.log10(size / 2) / Math.log10(localWorkSize * 2));
        
        OpenCL.clSetKernelArg(kernel, 1, 1);
        
        int stage = 1;
        int calls = 0;
        for(int i = 0; i < log && s > 0; ++i)
        {
            workSize.put(0, s);
            
            OpenCL.clEnqueueNDRangeKernel(queue, kernel, 1, null, workSize, workSizeLocal, null, null);
            
            stage *= 2 * workSizeLocal.get(0);
            
            OpenCL.clSetKernelArg(kernel, 1, stage);
            
            s /= 2 * workSizeLocal.get(0);

            workSizeLocal.put(0, s <= workSizeLocal.get(0) ? s : workSizeLocal.get(0));
            
            calls++;
        }
        
        OpenCL.clEnqueueReadBuffer(queue, input, 1, 0, data, null, null);

        int sum = (size * size + size) / 2;
        System.out.println(String.format("    cpu sum(1, %d) = %d", size, sum));
        System.out.println(String.format("    gpu sum(1, %d) = %d s==0 %b, kernel calls=%d, log=%d\n--- Test end.\n", size, data.get(0), (s==0), calls, log));

        OpenCL.clReleaseMemObject(input);
    }
    
    public static void runTest11(int size, CLKernel k)
    {
        IntBuffer data = BufferUtils.createIntBuffer(size);
        for(int i = 0; i < size; ++i)
        {
            data.put(1 + i);
        }
        
        data.position(0);
        
        CLMem d = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_WRITE | CL10.CL_MEM_COPY_HOST_PTR, data);
        OpenCL.clSetKernelArg(k, 0, d);
        OpenCL.clSetKernelArg(k, 1, 1);
        OpenCL.clSetKernelArg(k, 2, size % 2);
        
        int stride = 1;
        PointerBuffer gws = new PointerBuffer(1);
        
        gws.put(0, getEven(size) / 2);
 
        for(int i = size / 2; i > 0; i >>= 1, stride <<= 1)
        {
            OpenCL.clSetKernelArg(k, 1, stride);
            
            OpenCL.clEnqueueNDRangeKernel(queue, k, 1, null, gws, null, null, null);;
            
            int odd = ((int)gws.get(0) / 2) % 2;
            
            OpenCL.clSetKernelArg(k, 2, odd);
            
            gws.put(0, getEven((int)gws.get(0)) / 2);
        }
        
        OpenCL.clEnqueueReadBuffer(queue, d, 1, 0, data, null, null);
        
        OpenCL.clReleaseMemObject(d);
        
        System.out.println(data.get(0));
        
        data = null;

    }
    
    private static int getEven(int number)
    {
        return number - number % 2;
    }
    
    public static void main(String[] args)
    {
        
        CLUtil.createCL();
        
        PlatformDevicePair pair = CLUtil.choosePlatformAndDevice();
        
        CLUtil.printDeviceInfos(pair.device);
        
        context = OpenCL.clCreateContext(pair.platform, pair.device, null, null);
        queue = OpenCL.clCreateCommandQueue(context, pair.device, 0);
        program = OpenCL.clCreateProgramWithSource(context, IOUtil.readFileContent("kernel/reduction.cl"));
        OpenCL.clBuildProgram(program, pair.device, "", null);
        kernel = OpenCL.clCreateKernel(program, "reduce2");
        CLKernel k = OpenCL.clCreateKernel(program, "reduce3");
        
        System.out.println("\n--- Reduction Testing ---\n\n");
        
        for(int i = 7; i < 15; ++i)
        {
            runTest13(i, 8); //variant 1
        } 

        runTest13(2, 32);
        
        for(int i = 1; i < 1024; ++i)
        {
           // runTest11(i, k); //variant 3
        }
               
        OpenCL.clReleaseKernel(k);
        OpenCL.clReleaseCommandQueue(queue);
        OpenCL.clReleaseKernel(kernel);
        OpenCL.clReleaseProgram(program);
        OpenCL.clReleaseContext(context);
        
        CLUtil.destroyCL();
    }
}
